@@ -18,6 +18,7 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
+#include "config.h"
 #include "config_csync.h"
 
 #ifndef _GNU_SOURCE
@@ -26,23 +27,45 @@
 
 #include "c_lib.h"
 #include "c_private.h"
-#include "c_utf8.h"
 
-#include "csync_private.h"
 #include "csync_exclude.h"
-#include "csync_misc.h"
 
+#include "../version.h"
+#include "common/filesystembase.h"
 #include "common/utility.h"
 
 #include <QString>
 #include <QFileInfo>
+#include <QFile>
+#include <QDir>
 
+namespace {
+
+// See http://support.microsoft.com/kb/74496 and
+// https://msdn.microsoft.com/en-us/library/windows/desktop/aa365247(v=vs.85).aspx
+// Additionally, we ignore '$Recycle.Bin', see https://github.com/owncloud/client/issues/2955
+const QLatin1String win_device_names[] = {
+    QLatin1String("CON"), QLatin1String("PRN"), QLatin1String("AUX"), QLatin1String("NUL"),
+    QLatin1String("COM1"), QLatin1String("COM2"), QLatin1String("COM3"),
+    QLatin1String("COM4"), QLatin1String("COM5"), QLatin1String("COM6"),
+    QLatin1String("COM7"), QLatin1String("COM8"), QLatin1String("COM9"),
+    QLatin1String("LPT1"), QLatin1String("LPT2"), QLatin1String("LPT3"),
+    QLatin1String("LPT4"), QLatin1String("LPT5"), QLatin1String("LPT6"),
+    QLatin1String("LPT7"), QLatin1String("LPT8"), QLatin1String("LPT9"),
+    QLatin1String("CLOCK$")
+};
+
+const QLatin1String win_system_files[] = {
+    QLatin1String("$Recycle.Bin"),
+    QLatin1String("System Volume Information")
+};
+}
 
 /** Expands C-like escape sequences (in place)
  */
-static void csync_exclude_expand_escapes(QByteArray &input)
+OCSYNC_EXPORT void csync_exclude_expand_escapes(QByteArray &input)
 {
-    size_t o = 0;
+    qsizetype o = 0;
     char *line = input.data();
     auto len = input.size();
     for (int i = 0; i < len; ++i) {
@@ -73,109 +96,89 @@ static void csync_exclude_expand_escapes(QByteArray &input)
             line[o++] = line[i];
         }
     }
-    input.resize(OCC::Utility::convertSizeToUint(o));
+    input.resize(o);
 }
 
-// See http://support.microsoft.com/kb/74496 and
-// https://msdn.microsoft.com/en-us/library/windows/desktop/aa365247(v=vs.85).aspx
-// Additionally, we ignore '$Recycle.Bin', see https://github.com/owncloud/client/issues/2955
-static const char *win_reserved_words_3[] = { "CON", "PRN", "AUX", "NUL" };
-static const char *win_reserved_words_4[] = {
-    "COM1", "COM2", "COM3", "COM4", "COM5", "COM6", "COM7", "COM8", "COM9",
-    "LPT1", "LPT2", "LPT3", "LPT4", "LPT5", "LPT6", "LPT7", "LPT8", "LPT9"
-};
-static const char *win_reserved_words_n[] = { "CLOCK$", "$Recycle.Bin" };
-
 /**
  * @brief Checks if filename is considered reserved by Windows
  * @param file_name filename
  * @return true if file is reserved, false otherwise
  */
-bool csync_is_windows_reserved_word(const char *filename)
+OCSYNC_EXPORT bool csync_is_windows_reserved_word(const QStringRef &filename)
 {
-    size_t len_filename = strlen(filename);
+    size_t len_filename = filename.size();
 
     // Drive letters
-    if (len_filename == 2 && filename[1] == ':') {
-        if (filename[0] >= 'a' && filename[0] <= 'z') {
+    if (len_filename == 2 && filename.at(1) == QLatin1Char(':')) {
+        if (filename.at(0) >= QLatin1Char('a') && filename.at(0) <= QLatin1Char('z')) {
             return true;
         }
-        if (filename[0] >= 'A' && filename[0] <= 'Z') {
+        if (filename.at(0) >= QLatin1Char('A') && filename.at(0) <= QLatin1Char('Z')) {
             return true;
         }
     }
 
-    if (len_filename == 3 || (len_filename > 3 && filename[3] == '.')) {
-        for (const char *word : win_reserved_words_3) {
-            if (c_strncasecmp(filename, word, 3) == 0) {
-                return true;
-            }
+    // win_reserved_words is ordered so we can stop once we encounter a longer word
+    Q_ASSERT(std::is_sorted(std::begin(win_device_names), std::end(win_device_names), [](const QLatin1String &a, const QLatin1String &b) {
+        return a.size() < b.size();
+    }));
+    for (const auto &word : win_device_names) {
+        if (static_cast<size_t>(word.size()) > len_filename) {
+            break;
         }
+        // until windows 11, not only the device names where illegal file names
+        // also COM9.png was illegal
+        if ((static_cast<size_t>(word.size()) == len_filename || filename.at(word.size()) == QLatin1Char('.')) && filename.startsWith(word, Qt::CaseInsensitive)) {
+            return true;
+        }
     }
-
-    if (len_filename == 4 || (len_filename > 4 && filename[4] == '.')) {
-        for (const char *word : win_reserved_words_4) {
-            if (c_strncasecmp(filename, word, 4) == 0) {
-                return true;
-            }
+    // win_reserved_words is ordered so we can stop once we encounter a longer word
+    Q_ASSERT(std::is_sorted(std::begin(win_system_files), std::end(win_system_files), [](const QLatin1String &a, const QLatin1String &b) {
+        return a.size() < b.size();
+    }));
+    for (const auto &word : win_system_files) {
+        if (static_cast<size_t>(word.size()) > len_filename) {
+            break;
         }
-    }
-
-    for (const char *word : win_reserved_words_n) {
-        size_t len_word = strlen(word);
-        if (len_word == len_filename && c_strncasecmp(filename, word, len_word) == 0) {
+        if (static_cast<size_t>(word.size()) == len_filename && filename.compare(word, Qt::CaseInsensitive) == 0) {
             return true;
         }
     }
-
     return false;
 }
 
-static CSYNC_EXCLUDE_TYPE _csync_excluded_common(const char *path, bool excludeConflictFiles)
+static CSYNC_EXCLUDE_TYPE _csync_excluded_common(const QStringRef &path, bool excludeConflictFiles)
 {
-    const char *bname = nullptr;
-    size_t blen = 0;
-    int rc = -1;
-    CSYNC_EXCLUDE_TYPE match = CSYNC_NOT_EXCLUDED;
-
     /* split up the path */
-    bname = strrchr(path, '/');
-    if (bname) {
-        bname += 1; // don't include the /
-    } else {
-        bname = path;
+    QStringRef bname(path);
+    int lastSlash = path.lastIndexOf(QLatin1Char('/'));
+    if (lastSlash >= 0) {
+        bname = path.mid(lastSlash + 1);
     }
-    blen = strlen(bname);
 
+    qsizetype blen = bname.size();
     // 9 = strlen(".sync_.db")
-    if (blen >= 9 && bname[0] == '.') {
-        rc = csync_fnmatch("._sync_*.db*", bname, 0);
-        if (rc == 0) {
-            match = CSYNC_FILE_SILENTLY_EXCLUDED;
-            goto out;
+    if (blen >= 9 && bname.at(0) == QLatin1Char('.')) {
+        if (bname.contains(QLatin1String(".db"))) {
+            if (bname.startsWith(QLatin1String("._sync_"), Qt::CaseInsensitive)  // "._sync_*.db*"
+                || bname.startsWith(QLatin1String(".sync_"), Qt::CaseInsensitive) // ".sync_*.db*"
+                || bname.startsWith(QLatin1String(".csync_journal.db"), Qt::CaseInsensitive)) { // ".csync_journal.db*"
+                return CSYNC_FILE_SILENTLY_EXCLUDED;
+            }
         }
-        rc = csync_fnmatch(".sync_*.db*", bname, 0);
-        if (rc == 0) {
-            match = CSYNC_FILE_SILENTLY_EXCLUDED;
-            goto out;
-        }
-        rc = csync_fnmatch(".csync_journal.db*", bname, 0);
-        if (rc == 0) {
-            match = CSYNC_FILE_SILENTLY_EXCLUDED;
-            goto out;
-        }
-        rc = csync_fnmatch(".owncloudsync.log*", bname, 0);
-        if (rc == 0) {
-            match = CSYNC_FILE_SILENTLY_EXCLUDED;
-            goto out;
+        if (bname.startsWith(QLatin1String(".owncloudsync.log"), Qt::CaseInsensitive)) { // ".owncloudsync.log*"
+            return CSYNC_FILE_SILENTLY_EXCLUDED;
         }
     }
 
-    // check the strlen and ignore the file if its name is longer than 254 chars.
+    if (bname.endsWith(QLatin1String(APPLICATION_DOTVIRTUALFILE_SUFFIX), Qt::CaseInsensitive)) { // ".owncloud" placeholder
+        return CSYNC_FILE_EXCLUDE_RESERVED;
+    }
+
+    // check the strlen and ignore the file if its name is longer than 148 chars.
     // whenever changing this also check createDownloadTmpFileName
     if (blen > 148) {
-        match = CSYNC_FILE_EXCLUDE_LONG_FILENAME;
-        goto out;
+        return CSYNC_FILE_EXCLUDE_LONG_FILENAME;
     }
 
 #ifdef _WIN32
@@ -184,23 +187,25 @@ static CSYNC_EXCLUDE_TYPE _csync_excluded_common(const char *path, bool excludeC
     // as '.' is a separator that is not stored internally, so let's
     // not allow to sync those to avoid file loss/ambiguities (#416)
     if (blen > 1) {
-        if (bname[blen-1]== ' ') {
-            match = CSYNC_FILE_EXCLUDE_TRAILING_SPACE;
-            goto out;
-        } else if (bname[blen-1]== '.' ) {
-            match = CSYNC_FILE_EXCLUDE_INVALID_CHAR;
-            goto out;
+        if (bname.at(blen - 1) == QLatin1Char(' ')) {
+            return CSYNC_FILE_EXCLUDE_TRAILING_SPACE;
+        } else if (bname.at(blen - 1) == QLatin1Char('.')) {
+            return CSYNC_FILE_EXCLUDE_INVALID_CHAR;
         }
     }
 
     if (csync_is_windows_reserved_word(bname)) {
-      match = CSYNC_FILE_EXCLUDE_INVALID_CHAR;
-      goto out;
+        return CSYNC_FILE_EXCLUDE_INVALID_CHAR;
     }
 
     // Filter out characters not allowed in a filename on windows
-    for (const char *p = path; *p; p++) {
-        switch (*p) {
+    // see https://docs.microsoft.com/en-us/windows/desktop/fileio/naming-a-file
+    for (auto p : path) {
+        ushort c = p.unicode();
+        if (c < 32) {
+            return CSYNC_FILE_EXCLUDE_INVALID_CHAR;
+        }
+        switch (c) {
         case '\\':
         case ':':
         case '?':
@@ -209,69 +214,43 @@ static CSYNC_EXCLUDE_TYPE _csync_excluded_common(const char *path, bool excludeC
         case '>':
         case '<':
         case '|':
-            match = CSYNC_FILE_EXCLUDE_INVALID_CHAR;
-            goto out;
+            return CSYNC_FILE_EXCLUDE_INVALID_CHAR;
         default:
             break;
         }
     }
 #endif
 
-    /* We create a Desktop.ini on Windows for the sidebar icon, make sure we don't sync it. */
-    if (blen == 11 && path == bname) {
-        rc = csync_fnmatch("Desktop.ini", bname, 0);
-        if (rc == 0) {
-            match = CSYNC_FILE_SILENTLY_EXCLUDED;
-            goto out;
-        }
+    /* Do not sync desktop.ini files anywhere in the tree. */
+    const auto desktopIniFile = QStringLiteral("desktop.ini");
+    if (blen == static_cast<qsizetype>(desktopIniFile.length()) && bname.compare(desktopIniFile, Qt::CaseInsensitive) == 0) {
+        return CSYNC_FILE_SILENTLY_EXCLUDED;
     }
 
-    if (excludeConflictFiles && OCC::Utility::isConflictFile(bname)) {
-        match = CSYNC_FILE_EXCLUDE_CONFLICT;
-        goto out;
+
+    if (excludeConflictFiles && OCC::Utility::isConflictFile(path)) {
+        return CSYNC_FILE_EXCLUDE_CONFLICT;
     }
-
-  out:
-
-    return match;
+    return CSYNC_NOT_EXCLUDED;
 }
 
-static QByteArray leftIncludeLast(const QByteArray & arr, char c)
-{
-    // left up to and including `c`
-    return arr.left(arr.lastIndexOf(c, arr.size() - 2) + 1);
-}
 
 using namespace OCC;
 
-ExcludedFiles::ExcludedFiles(QString localPath)
-    : _localPath(std::move(localPath))
+ExcludedFiles::ExcludedFiles()
+    : _clientVersion(MIRALL_VERSION_MAJOR, MIRALL_VERSION_MINOR, MIRALL_VERSION_PATCH)
 {
-    Q_ASSERT(_localPath.endsWith("/"));
     // Windows used to use PathMatchSpec which allows *foo to match abc/deffoo.
     _wildcardsMatchSlash = Utility::isWindows();
-
-    // We're in a detached exclude probably coming from a partial sync or test
-    if (_localPath.isEmpty())
-        return;
-
-    // Load exclude file from base dir
-    QFileInfo fi(_localPath + ".sync-exclude.lst");
-    if (fi.isReadable())
-        addInTreeExcludeFilePath(fi.absoluteFilePath());
 }
 
-ExcludedFiles::~ExcludedFiles() = default;
+ExcludedFiles::~ExcludedFiles()
+{
+}
 
 void ExcludedFiles::addExcludeFilePath(const QString &path)
 {
-    _excludeFiles[_localPath.toUtf8()].append(path);
-}
-
-void ExcludedFiles::addInTreeExcludeFilePath(const QString &path)
-{
-    BasePathByteArray basePath = leftIncludeLast(path.toUtf8(), '/');
-    _excludeFiles[basePath].append(path);
+    _excludeFiles.insert(path);
 }
 
 void ExcludedFiles::setExcludeConflictFiles(bool onoff)
@@ -279,24 +258,11 @@ void ExcludedFiles::setExcludeConflictFiles(bool onoff)
     _excludeConflictFiles = onoff;
 }
 
-void ExcludedFiles::addManualExclude(const QByteArray &expr)
+void ExcludedFiles::addManualExclude(const QString &expr)
 {
-    addManualExclude(expr, _localPath.toUtf8());
-}
-
-void ExcludedFiles::addManualExclude(const QByteArray &expr, const QByteArray &basePath)
-{
-#if defined(Q_OS_WIN)
-    Q_ASSERT(basePath.size() >= 2 && basePath.at(1) == ':');
-#else
-    Q_ASSERT(basePath.startsWith('/'));
-#endif
-    Q_ASSERT(basePath.endsWith('/'));
-
-    auto key = basePath;
-    _manualExcludes[key].append(expr);
-    _allExcludes[key].append(expr);
-    prepare(key);
+    _manualExcludes.append(expr);
+    _allExcludes.append(expr);
+    prepare();
 }
 
 void ExcludedFiles::clearManualExcludes()
@@ -311,97 +277,123 @@ void ExcludedFiles::setWildcardsMatchSlash(bool onoff)
     prepare();
 }
 
-bool ExcludedFiles::loadExcludeFile(const QByteArray & basePath, const QString & file)
+void ExcludedFiles::setClientVersion(ExcludedFiles::Version version)
 {
-    QFile f(file);
-    if (!f.open(QIODevice::ReadOnly))
-        return false;
-
-    while (!f.atEnd()) {
-        QByteArray line = f.readLine().trimmed();
-        if (line.isEmpty() || line.startsWith('#'))
-            continue;
-        csync_exclude_expand_escapes(line);
-        _allExcludes[basePath].append(line);
-    }
-
-    // nothing to prepare if the user decided to not exclude anything
-    if(_allExcludes.size())
-        prepare(basePath);
-
-    return true;
+    _clientVersion = version;
 }
 
 bool ExcludedFiles::reloadExcludeFiles()
 {
     _allExcludes.clear();
-    // clear all regex
-    _bnameTraversalRegexFile.clear();
-    _bnameTraversalRegexDir.clear();
-    _fullTraversalRegexFile.clear();
-    _fullTraversalRegexDir.clear();
-    _fullRegexFile.clear();
-    _fullRegexDir.clear();
-
     bool success = true;
-    const auto keys = _excludeFiles.keys();
-    for (const auto& basePath : keys) {
-        for (const auto& file : _excludeFiles.value(basePath)) {
-            success = loadExcludeFile(basePath, file);
+    for (const auto &file : qAsConst(_excludeFiles)) {
+        QFile f(file);
+        if (!f.open(QIODevice::ReadOnly)) {
+            success = false;
+            continue;
+        }
+        while (!f.atEnd()) {
+            QByteArray line = f.readLine().trimmed();
+            if (line.startsWith("#!version")) {
+                if (!versionDirectiveKeepNextLine(line))
+                    f.readLine();
+            }
+            if (line.isEmpty() || line.startsWith('#'))
+                continue;
+            csync_exclude_expand_escapes(line);
+            _allExcludes.append(QString::fromUtf8(line));
         }
     }
-
-    auto endManual = _manualExcludes.cend();
-    for (auto kv = _manualExcludes.cbegin(); kv != endManual; ++kv) {
-        _allExcludes[kv.key()].append(kv.value());
-        prepare(kv.key());
-    }
-
+    _allExcludes.append(_manualExcludes);
+    prepare();
     return success;
 }
 
+bool ExcludedFiles::versionDirectiveKeepNextLine(const QByteArray &directive) const
+{
+    if (!directive.startsWith("#!version"))
+        return true;
+    QByteArrayList args = directive.split(' ');
+    if (args.size() != 3)
+        return true;
+    QByteArray op = args[1];
+    QByteArrayList argVersions = args[2].split('.');
+    if (argVersions.size() != 3)
+        return true;
+
+    auto argVersion = std::make_tuple(argVersions[0].toInt(), argVersions[1].toInt(), argVersions[2].toInt());
+    if (op == "<=")
+        return _clientVersion <= argVersion;
+    if (op == "<")
+        return _clientVersion < argVersion;
+    if (op == ">")
+        return _clientVersion > argVersion;
+    if (op == ">=")
+        return _clientVersion >= argVersion;
+    if (op == "==")
+        return _clientVersion == argVersion;
+    return true;
+}
+
 bool ExcludedFiles::isExcluded(
     const QString &filePath,
     const QString &basePath,
     bool excludeHidden) const
 {
-    if (!filePath.startsWith(basePath, Utility::fsCasePreserving() ? Qt::CaseInsensitive : Qt::CaseSensitive)) {
+    if (!FileSystem::isChildPathOf(filePath, basePath)) {
         // Mark paths we're not responsible for as excluded...
         return true;
     }
 
-    //TODO this seems a waste, hidden files are ignored before hitting this function it seems
+    const QFileInfo fileInfo(filePath);
+    if (!fileInfo.exists()) {
+        return isExcluded(fileInfo.path(), basePath, excludeHidden);
+    }
+
     if (excludeHidden) {
-        QString path = filePath;
+        QFileInfo fi = fileInfo;
         // Check all path subcomponents, but to *not* check the base path:
         // We do want to be able to sync with a hidden folder as the target.
-        while (path.size() > basePath.size()) {
-            QFileInfo fi(path);
-            if (fi.fileName() != ".sync-exclude.lst"
-                && (fi.isHidden() || fi.fileName().startsWith(QLatin1Char('.')))) {
+        while (fi.filePath().size() > basePath.size()) {
+            if (fi.isHidden()) {
                 return true;
             }
 
             // Get the parent path
-            path = fi.absolutePath();
+            fi = { fi.path() };
         }
     }
-
-    QFileInfo fi(filePath);
     ItemType type = ItemTypeFile;
-    if (fi.isDir()) {
+    if (fileInfo.isDir()) {
         type = ItemTypeDirectory;
     }
+    return isExcludedRemote(filePath, basePath, excludeHidden, type);
+}
 
-    QString relativePath = filePath.mid(basePath.size());
+bool ExcludedFiles::isExcludedRemote(const QString &filePath, const QString &basePath, bool excludeHidden, ItemType type) const
+{
+    if (!FileSystem::isChildPathOf(filePath, basePath)) {
+        // Mark paths we're not responsible for as excluded...
+        return true;
+    }
+
+    auto relativePath = filePath.midRef(basePath.size());
     if (relativePath.endsWith(QLatin1Char('/'))) {
         relativePath.chop(1);
     }
 
-    return fullPatternMatch(relativePath.toUtf8(), type) != CSYNC_NOT_EXCLUDED;
+    if (excludeHidden) {
+        // Check their parent directories for . hidden files
+        if (relativePath.startsWith(QLatin1Char('.')) || relativePath.contains(QLatin1String("/."))) {
+            return true;
+        }
+    }
+
+
+    return fullPatternMatch(relativePath, type) != CSYNC_NOT_EXCLUDED;
 }
 
-CSYNC_EXCLUDE_TYPE ExcludedFiles::traversalPatternMatch(const char *path, ItemType filetype)
+CSYNC_EXCLUDE_TYPE ExcludedFiles::traversalPatternMatch(const QStringRef &path, ItemType filetype) const
 {
     auto match = _csync_excluded_common(path, _excludeConflictFiles);
     if (match != CSYNC_NOT_EXCLUDED)
@@ -409,128 +401,76 @@ CSYNC_EXCLUDE_TYPE ExcludedFiles::traversalPatternMatch(const char *path, ItemTy
     if (_allExcludes.isEmpty())
         return CSYNC_NOT_EXCLUDED;
 
-    // Directories are guaranteed to be visited before their files
-    if (filetype == ItemTypeDirectory) {
-        QFileInfo fi = QFileInfo(_localPath + path + "/.sync-exclude.lst");
-        if (fi.isReadable()) {
-            addInTreeExcludeFilePath(fi.absoluteFilePath());
-            loadExcludeFile(fi.absolutePath().toUtf8(), fi.absoluteFilePath());
-        }
-    }
-
     // Check the bname part of the path to see whether the full
     // regex should be run.
-
-    const char *bname = strrchr(path, '/');
-    if (bname) {
-        bname += 1; // don't include the /
-    } else {
-        bname = path;
+    QStringRef bnameStr(path);
+    int lastSlash = path.lastIndexOf(QLatin1Char('/'));
+    if (lastSlash >= 0) {
+        bnameStr = path.mid(lastSlash + 1);
     }
-    QString bnameStr = QString::fromUtf8(bname);
 
-    QByteArray basePath(_localPath.toUtf8() + path);
-    while (basePath.size() > _localPath.size()) {
-        basePath = leftIncludeLast(basePath, '/');
-        QRegularExpressionMatch m;
-        if (filetype == ItemTypeDirectory
-            && _bnameTraversalRegexDir.contains(basePath)) {
-            m = _bnameTraversalRegexDir[basePath].match(bnameStr);
-        } else if (filetype == ItemTypeFile
-            && _bnameTraversalRegexFile.contains(basePath)) {
-            m = _bnameTraversalRegexFile[basePath].match(bnameStr);
-        } else {
-            continue;
-        }
+    QRegularExpressionMatch m;
+    if (filetype == ItemTypeDirectory) {
+        m = _bnameTraversalRegexDir.match(bnameStr);
+    } else {
+        m = _bnameTraversalRegexFile.match(bnameStr);
+    }
+    if (!m.hasMatch())
+        return CSYNC_NOT_EXCLUDED;
+    if (m.capturedStart(QStringLiteral("exclude")) != -1) {
+        return CSYNC_FILE_EXCLUDE_LIST;
+    } else if (m.capturedStart(QStringLiteral("excluderemove")) != -1) {
+        return CSYNC_FILE_EXCLUDE_AND_REMOVE;
+    }
 
-        if (!m.hasMatch())
-            return CSYNC_NOT_EXCLUDED;
+    // third capture: full path matching is triggered
+    QStringRef pathStr = path;
+
+    if (filetype == ItemTypeDirectory) {
+        m = _fullTraversalRegexDir.match(pathStr);
+    } else {
+        m = _fullTraversalRegexFile.match(pathStr);
+    }
+    if (m.hasMatch()) {
         if (m.capturedStart(QStringLiteral("exclude")) != -1) {
             return CSYNC_FILE_EXCLUDE_LIST;
         } else if (m.capturedStart(QStringLiteral("excluderemove")) != -1) {
             return CSYNC_FILE_EXCLUDE_AND_REMOVE;
         }
     }
-
-    // third capture: full path matching is triggered
-    QString pathStr = QString::fromUtf8(path);
-    basePath = _localPath.toUtf8() + path;
-    while (basePath.size() > _localPath.size()) {
-        basePath = leftIncludeLast(basePath, '/');
-        QRegularExpressionMatch m;
-        if (filetype == ItemTypeDirectory
-            && _fullTraversalRegexDir.contains(basePath)) {
-            m = _fullTraversalRegexDir[basePath].match(pathStr);
-        } else if (filetype == ItemTypeFile
-            && _fullTraversalRegexFile.contains(basePath)) {
-            m = _fullTraversalRegexFile[basePath].match(pathStr);
-        } else {
-            continue;
-        }
-
-        if (m.hasMatch()) {
-            if (m.capturedStart(QStringLiteral("exclude")) != -1) {
-                return CSYNC_FILE_EXCLUDE_LIST;
-            } else if (m.capturedStart(QStringLiteral("excluderemove")) != -1) {
-                return CSYNC_FILE_EXCLUDE_AND_REMOVE;
-            }
-        }
-    }
     return CSYNC_NOT_EXCLUDED;
 }
 
-CSYNC_EXCLUDE_TYPE ExcludedFiles::fullPatternMatch(const char *path, ItemType filetype) const
+CSYNC_EXCLUDE_TYPE ExcludedFiles::fullPatternMatch(const QStringRef &p, ItemType filetype) const
 {
-    auto match = _csync_excluded_common(path, _excludeConflictFiles);
+    auto match = _csync_excluded_common(p, _excludeConflictFiles);
     if (match != CSYNC_NOT_EXCLUDED)
         return match;
     if (_allExcludes.isEmpty())
         return CSYNC_NOT_EXCLUDED;
 
-    QString p = QString::fromUtf8(path);
-    // `path` seems to always be relative to `_localPath`, the tests however have not been
-    // written that way... this makes the tests happy for now. TODO Fix the tests at some point
-    if (path[0] == '/')
-        ++path;
-
-    QByteArray basePath(_localPath.toUtf8() + path);
-    while (basePath.size() > _localPath.size()) {
-        basePath = leftIncludeLast(basePath, '/');
-        QRegularExpressionMatch m;
-        if (filetype == ItemTypeDirectory
-            && _fullRegexDir.contains(basePath)) {
-            m = _fullRegexDir[basePath].match(p);
-        } else if (filetype == ItemTypeFile
-            && _fullRegexFile.contains(basePath)) {
-            m = _fullRegexFile[basePath].match(p);
-        } else {
-            continue;
-        }
-
-        if (m.hasMatch()) {
-            if (m.capturedStart(QStringLiteral("exclude")) != -1) {
-                return CSYNC_FILE_EXCLUDE_LIST;
-            } else if (m.capturedStart(QStringLiteral("excluderemove")) != -1) {
-                return CSYNC_FILE_EXCLUDE_AND_REMOVE;
-            }
+    QRegularExpressionMatch m;
+    if (filetype == ItemTypeDirectory) {
+        m = _fullRegexDir.match(p);
+    } else {
+        m = _fullRegexFile.match(p);
+    }
+    if (m.hasMatch()) {
+        if (m.capturedStart(QStringLiteral("exclude")) != -1) {
+            return CSYNC_FILE_EXCLUDE_LIST;
+        } else if (m.capturedStart(QStringLiteral("excluderemove")) != -1) {
+            return CSYNC_FILE_EXCLUDE_AND_REMOVE;
         }
     }
-
     return CSYNC_NOT_EXCLUDED;
 }
 
-auto ExcludedFiles::csyncTraversalMatchFun()
-    -> std::function<CSYNC_EXCLUDE_TYPE(const char *path, ItemType filetype)>
-{
-    return [this](const char *path, ItemType filetype) { return this->traversalPatternMatch(path, filetype); };
-}
-
 /**
  * On linux we used to use fnmatch with FNM_PATHNAME, but the windows function we used
  * didn't have that behavior. wildcardsMatchSlash can be used to control which behavior
  * the resulting regex shall use.
  */
-static QString convertToRegexpSyntax(QString exclude, bool wildcardsMatchSlash)
+QString ExcludedFiles::convertToRegexpSyntax(QString exclude, bool wildcardsMatchSlash)
 {
     // Translate *, ?, [...] to their regex variants.
     // The escape sequences \*, \?, \[. \\ have a special meaning,
@@ -555,17 +495,17 @@ static QString convertToRegexpSyntax(QString exclude, bool wildcardsMatchSlash)
         case '*':
             flush();
             if (wildcardsMatchSlash) {
-                regex.append(".*");
+                regex.append(QLatin1String(".*"));
             } else {
-                regex.append("[^/]*");
+                regex.append(QLatin1String("[^/]*"));
             }
             break;
         case '?':
             flush();
             if (wildcardsMatchSlash) {
-                regex.append(".");
+                regex.append(QLatin1Char('.'));
             } else {
-                regex.append("[^/]");
+                regex.append(QStringLiteral("[^/]"));
             }
             break;
         case '[': {
@@ -573,20 +513,20 @@ static QString convertToRegexpSyntax(QString exclude, bool wildcardsMatchSlash)
             // Find the end of the bracket expression
             auto j = i + 1;
             for (; j < len; ++j) {
-                if (exclude[j] == ']')
+                if (exclude[j] == QLatin1Char(']'))
                     break;
-                if (j != len - 1 && exclude[j] == '\\' && exclude[j + 1] == ']')
+                if (j != len - 1 && exclude[j] == QLatin1Char('\\') && exclude[j + 1] == QLatin1Char(']'))
                     ++j;
             }
             if (j == len) {
                 // no matching ], just insert the escaped [
-                regex.append("\\[");
+                regex.append(QStringLiteral("\\["));
                 break;
             }
             // Translate [! to [^
             QString bracketExpr = exclude.mid(i, j - i + 1);
-            if (bracketExpr.startsWith("[!"))
-                bracketExpr[1] = '^';
+            if (bracketExpr.startsWith(QLatin1String("[!")))
+                bracketExpr[1] = QLatin1Char('^');
             regex.append(bracketExpr);
             i = j;
             break;
@@ -594,7 +534,7 @@ static QString convertToRegexpSyntax(QString exclude, bool wildcardsMatchSlash)
         case '\\':
             flush();
             if (i == len - 1) {
-                regex.append("\\\\");
+                regex.append(QStringLiteral("\\\\"));
                 break;
             }
             // '\*' -> '\*', but '\z' -> '\\z'
@@ -620,11 +560,11 @@ static QString convertToRegexpSyntax(QString exclude, bool wildcardsMatchSlash)
     return regex;
 }
 
-static QString extractBnameTrigger(const QString &exclude, bool wildcardsMatchSlash)
+QString ExcludedFiles::extractBnameTrigger(const QString &exclude, bool wildcardsMatchSlash)
 {
     // We can definitely drop everything to the left of a / - that will never match
     // any bname.
-    QString pattern = exclude.mid(exclude.lastIndexOf('/') + 1);
+    QString pattern = exclude.mid(exclude.lastIndexOf(QLatin1Char('/')) + 1);
 
     // Easy case, nothing else can match a slash, so that's it.
     if (!wildcardsMatchSlash)
@@ -651,30 +591,13 @@ static QString extractBnameTrigger(const QString &exclude, bool wildcardsMatchSl
 
     // And if there was a wildcard, it starts with a *
     if (i >= 0)
-        pattern.prepend('*');
+        pattern.prepend(QLatin1Char('*'));
 
     return pattern;
 }
 
 void ExcludedFiles::prepare()
 {
-    // clear all regex
-    _bnameTraversalRegexFile.clear();
-    _bnameTraversalRegexDir.clear();
-    _fullTraversalRegexFile.clear();
-    _fullTraversalRegexDir.clear();
-    _fullRegexFile.clear();
-    _fullRegexDir.clear();
-
-    const auto keys = _allExcludes.keys();
-    for (auto const & basePath : keys)
-        prepare(basePath);
-}
-
-void ExcludedFiles::prepare(const BasePathByteArray & basePath)
-{
-    Q_ASSERT(_allExcludes.contains(basePath));
-
     // Build regular expressions for the different cases.
     //
     // To compose the _bnameTraversalRegex, _fullTraversalRegex and _fullRegex
@@ -712,25 +635,25 @@ void ExcludedFiles::prepare(const BasePathByteArray & basePath)
     auto regexAppend = [](QString &fileDirPattern, QString &dirPattern, const QString &appendMe, bool dirOnly) {
         QString &pattern = dirOnly ? dirPattern : fileDirPattern;
         if (!pattern.isEmpty())
-            pattern.append("|");
+            pattern.append(QLatin1Char('|'));
         pattern.append(appendMe);
     };
 
-    for (auto exclude : _allExcludes.value(basePath)) {
-        if (exclude[0] == '\n')
+    for (auto exclude : qAsConst(_allExcludes)) {
+        if (exclude[0] == QLatin1Char('\n'))
             continue; // empty line
-        if (exclude[0] == '\r')
+        if (exclude[0] == QLatin1Char('\r'))
             continue; // empty line
 
-        bool matchDirOnly = exclude.endsWith('/');
+        bool matchDirOnly = exclude.endsWith(QLatin1Char('/'));
         if (matchDirOnly)
             exclude = exclude.left(exclude.size() - 1);
 
-        bool removeExcluded = (exclude[0] == ']');
+        bool removeExcluded = (exclude[0] == QLatin1Char(']'));
         if (removeExcluded)
             exclude = exclude.mid(1);
 
-        bool fullPath = exclude.contains('/');
+        bool fullPath = exclude.contains(QLatin1Char('/'));
 
         /* Use QRegularExpression, append to the right pattern */
         auto &bnameFileDir = removeExcluded ? bnameFileDirRemove : bnameFileDirKeep;
@@ -738,16 +661,7 @@ void ExcludedFiles::prepare(const BasePathByteArray & basePath)
         auto &fullFileDir = removeExcluded ? fullFileDirRemove : fullFileDirKeep;
         auto &fullDir = removeExcluded ? fullDirRemove : fullDirKeep;
 
-        if (fullPath) {
-            // The full pattern is matched against a path relative to _localPath, however exclude is
-            // relative to basePath at this point.
-            // We know for sure that both _localPath and basePath are absolute and that basePath is
-            // contained in _localPath. So we can simply remove it from the begining.
-            auto relPath = basePath.mid(_localPath.size());
-            // Make exclude relative to _localPath
-            exclude.prepend(relPath);
-        }
-        auto regexExclude = convertToRegexpSyntax(QString::fromUtf8(exclude), _wildcardsMatchSlash);
+        auto regexExclude = convertToRegexpSyntax(exclude, _wildcardsMatchSlash);
         if (!fullPath) {
             regexAppend(bnameFileDir, bnameDir, regexExclude, matchDirOnly);
         } else {
@@ -763,7 +677,7 @@ void ExcludedFiles::prepare(const BasePathByteArray & basePath)
     // The empty pattern would match everything - change it to match-nothing
     auto emptyMatchNothing = [](QString &pattern) {
         if (pattern.isEmpty())
-            pattern = "a^";
+            pattern = QStringLiteral("a^");
     };
     emptyMatchNothing(fullFileDirKeep);
     emptyMatchNothing(fullFileDirRemove);
@@ -783,73 +697,73 @@ void ExcludedFiles::prepare(const BasePathByteArray & basePath)
     // (exclude)|(excluderemove)|(bname triggers).
     // If the third group matches, the fullActivatedRegex needs to be applied
     // to the full path.
-    _bnameTraversalRegexFile[basePath].setPattern(
-        "^(?P<exclude>" + bnameFileDirKeep + ")$|"
-        + "^(?P<excluderemove>" + bnameFileDirRemove + ")$|"
-        + "^(?P<trigger>" + bnameTriggerFileDir + ")$");
-    _bnameTraversalRegexDir[basePath].setPattern(
-        "^(?P<exclude>" + bnameFileDirKeep + "|" + bnameDirKeep + ")$|"
-        + "^(?P<excluderemove>" + bnameFileDirRemove + "|" + bnameDirRemove + ")$|"
-        + "^(?P<trigger>" + bnameTriggerFileDir + "|" + bnameTriggerDir + ")$");
+    _bnameTraversalRegexFile.setPattern(
+        QStringLiteral("^(?P<exclude>%1)$|"
+                       "^(?P<excluderemove>%2)$|"
+                       "^(?P<trigger>%3)$")
+            .arg(bnameFileDirKeep, bnameFileDirRemove, bnameTriggerFileDir));
+    _bnameTraversalRegexDir.setPattern(
+        QStringLiteral("^(?P<exclude>%1|%2)$|"
+                       "^(?P<excluderemove>%3|%4)$|"
+                       "^(?P<trigger>%5|%6)$")
+            .arg(bnameFileDirKeep, bnameDirKeep, bnameFileDirRemove, bnameDirRemove, bnameTriggerFileDir, bnameTriggerDir));
 
     // The full traveral regex is applied to the full path if the trigger capture of
     // the bname regex matches. Its basic form is (exclude)|(excluderemove)".
     // This pattern can be much simpler than fullRegex since we can assume a traversal
     // situation and doesn't need to look for bname patterns in parent paths.
-    _fullTraversalRegexFile[basePath].setPattern(
-        QLatin1String("")
+    _fullTraversalRegexFile.setPattern(
         // Full patterns are anchored to the beginning
-        + "^(?P<exclude>" + fullFileDirKeep + ")(?:$|/)"
-        + "|"
-        + "^(?P<excluderemove>" + fullFileDirRemove + ")(?:$|/)");
-    _fullTraversalRegexDir[basePath].setPattern(
-        QLatin1String("")
-        + "^(?P<exclude>" + fullFileDirKeep + "|" + fullDirKeep + ")(?:$|/)"
-        + "|"
-        + "^(?P<excluderemove>" + fullFileDirRemove + "|" + fullDirRemove + ")(?:$|/)");
+        QStringLiteral("^(?P<exclude>%1)(?:$|/)"
+                       "|"
+                       "^(?P<excluderemove>%2)(?:$|/)")
+            .arg(fullFileDirKeep, fullFileDirRemove));
+    _fullTraversalRegexDir.setPattern(
+        QStringLiteral("^(?P<exclude>%1|%2)(?:$|/)"
+                       "|"
+                       "^(?P<excluderemove>%3|%4)(?:$|/)")
+            .arg(fullFileDirKeep, fullDirKeep, fullFileDirRemove, fullDirRemove));
 
     // The full regex is applied to the full path and incorporates both bname and
     // full-path patterns. It has the form "(exclude)|(excluderemove)".
-    _fullRegexFile[basePath].setPattern(
-        QLatin1String("(?P<exclude>")
-        // Full patterns are anchored to the beginning
-        + "^(?:" + fullFileDirKeep + ")(?:$|/)" + "|"
-        // Simple bname patterns can be any path component
-        + "(?:^|/)(?:" + bnameFileDirKeep + ")(?:$|/)" + "|"
-        // When checking a file for exclusion we must check all parent paths
-        // against the dir-only patterns as well.
-        + "(?:^|/)(?:" + bnameDirKeep + ")/"
-        + ")"
-        + "|"
-        + "(?P<excluderemove>"
-        + "^(?:" + fullFileDirRemove + ")(?:$|/)" + "|"
-        + "(?:^|/)(?:" + bnameFileDirRemove + ")(?:$|/)" + "|"
-        + "(?:^|/)(?:" + bnameDirRemove + ")/"
-        + ")");
-    _fullRegexDir[basePath].setPattern(
-        QLatin1String("(?P<exclude>")
-        + "^(?:" + fullFileDirKeep + "|" + fullDirKeep + ")(?:$|/)" + "|"
-        + "(?:^|/)(?:" + bnameFileDirKeep + "|" + bnameDirKeep + ")(?:$|/)"
-        + ")"
-        + "|"
-        + "(?P<excluderemove>"
-        + "^(?:" + fullFileDirRemove + "|" + fullDirRemove + ")(?:$|/)" + "|"
-        + "(?:^|/)(?:" + bnameFileDirRemove + "|" + bnameDirRemove + ")(?:$|/)"
-        + ")");
+    _fullRegexFile.setPattern(
+        QStringLiteral("(?P<exclude>"
+                       // Full patterns are anchored to the beginning
+                       "^(?:%1)(?:$|/)|"
+                       // Simple bname patterns can be any path component
+                       "(?:^|/)(?:%2)(?:$|/)|"
+                       // When checking a file for exclusion we must check all parent paths
+                       // against the dir-only patterns as well.
+                       "(?:^|/)(?:%3)/)"
+                       "|"
+                       "(?P<excluderemove>"
+                       "^(?:%4)(?:$|/)|"
+                       "(?:^|/)(?:%5)(?:$|/)|"
+                       "(?:^|/)(?:%6)/)")
+            .arg(fullFileDirKeep, bnameFileDirKeep, bnameDirKeep, fullFileDirRemove, bnameFileDirRemove, bnameDirRemove));
+    _fullRegexDir.setPattern(
+        QStringLiteral("(?P<exclude>"
+                       "^(?:%1|%2)(?:$|/)|"
+                       "(?:^|/)(?:%3|%4)(?:$|/))"
+                       "|"
+                       "(?P<excluderemove>"
+                       "^(?:%5|%6)(?:$|/)|"
+                       "(?:^|/)(?:%7|%8)(?:$|/))")
+            .arg(fullFileDirKeep, fullDirKeep, bnameFileDirKeep, bnameDirKeep, fullFileDirRemove, fullDirRemove, bnameFileDirRemove, bnameDirRemove));
 
     QRegularExpression::PatternOptions patternOptions = QRegularExpression::NoPatternOption;
     if (OCC::Utility::fsCasePreserving())
         patternOptions |= QRegularExpression::CaseInsensitiveOption;
-    _bnameTraversalRegexFile[basePath].setPatternOptions(patternOptions);
-    _bnameTraversalRegexFile[basePath].optimize();
-    _bnameTraversalRegexDir[basePath].setPatternOptions(patternOptions);
-    _bnameTraversalRegexDir[basePath].optimize();
-    _fullTraversalRegexFile[basePath].setPatternOptions(patternOptions);
-    _fullTraversalRegexFile[basePath].optimize();
-    _fullTraversalRegexDir[basePath].setPatternOptions(patternOptions);
-    _fullTraversalRegexDir[basePath].optimize();
-    _fullRegexFile[basePath].setPatternOptions(patternOptions);
-    _fullRegexFile[basePath].optimize();
-    _fullRegexDir[basePath].setPatternOptions(patternOptions);
-    _fullRegexDir[basePath].optimize();
+    _bnameTraversalRegexFile.setPatternOptions(patternOptions);
+    _bnameTraversalRegexFile.optimize();
+    _bnameTraversalRegexDir.setPatternOptions(patternOptions);
+    _bnameTraversalRegexDir.optimize();
+    _fullTraversalRegexFile.setPatternOptions(patternOptions);
+    _fullTraversalRegexFile.optimize();
+    _fullTraversalRegexDir.setPatternOptions(patternOptions);
+    _fullTraversalRegexDir.optimize();
+    _fullRegexFile.setPatternOptions(patternOptions);
+    _fullRegexFile.optimize();
+    _fullRegexDir.setPatternOptions(patternOptions);
+    _fullRegexDir.optimize();
 }
