@@ -390,7 +390,8 @@ void OAuth::refreshAuthentication(const QString &refreshToken)
 void OAuth::finalize(const QPointer<QTcpSocket> &socket, const QString &accessToken,
     const QString &refreshToken, const QString &user, const QUrl &messageUrl)
 {
-    if (!_account->davUser().isNull() && !user.isNull() && user.toLower() != _account->davUser().toLower()) {
+    // dav user names are case insensetive, we might compare a user input with the string provided by the server
+    if (!_account->davUser().isEmpty() && user.compare(_account->davUser(), Qt::CaseInsensitive) != 0) {
         // Connected with the wrong user
         qCWarning(lcOauth) << "We expected the user" << _account->davUser() << "but the server answered with user" << user;
         const QString message = tr("<h1>Wrong user</h1>"
