@@ -314,12 +314,20 @@ LsColJob::LsColJob(AccountPtr account, const QString &path, QObject *parent)
     : AbstractNetworkJob(account, QString(), parent)
     , _url(makeDavUrl(path))
 {
+    // Always have a higher priority than the propagator because we use this from the UI
+    // and really want this to be done first (no matter what internal scheduling QNAM uses).
+    // Also possibly useful for avoiding false timeouts.
+    setPriority(QNetworkRequest::HighPriority);
 }
 
 LsColJob::LsColJob(AccountPtr account, const QUrl &url, QObject *parent)
     : AbstractNetworkJob(account, QString(), parent)
     , _url(url)
 {
+    // Always have a higher priority than the propagator because we use this from the UI
+    // and really want this to be done first (no matter what internal scheduling QNAM uses).
+    // Also possibly useful for avoiding false timeouts.
+    setPriority(QNetworkRequest::HighPriority);
 }
 
 void LsColJob::setProperties(const QList<QByteArray> &properties)
@@ -582,10 +590,6 @@ void PropfindJob::start()
         Q_EMIT result(values);
     });
     QNetworkRequest req;
-    // Always have a higher priority than the propagator because we use this from the UI
-    // and really want this to be done first (no matter what internal scheduling QNAM uses).
-    // Also possibly useful for avoiding false timeouts.
-    req.setPriority(QNetworkRequest::HighPriority);
     req.setRawHeader(QByteArrayLiteral("Depth"), QByteArrayLiteral("0"));
     startImpl(req);
 }
