@@ -84,6 +84,16 @@ State machine:
 
 void PropagateUploadFileNG::doStartUpload()
 {
+    const QString fileName = propagator()->fullLocalPath(_item->_file);
+    // If the file is currently locked, we want to retry the sync
+    // when it becomes available again.
+    const auto lockMode = propagator()->syncOptions().requiredLockMode();
+    if (FileSystem::isFileLocked(fileName, lockMode)) {
+        emit propagator()->seenLockedFile(fileName, lockMode);
+        abortWithError(SyncFileItem::SoftError, tr("%1 the file is currently in use").arg(QDir::toNativeSeparators(fileName)));
+        return;
+    }
+
     propagator()->_activeJobList.append(this);
 
     UploadRangeInfo rangeinfo = { 0, _item->_size };
@@ -348,6 +358,19 @@ void PropagateUploadFileNG::doFinalMove()
 
     const QUrl source = Utility::concatUrlPath(chunkUrl(), QStringLiteral("/.file"));
 
+#ifdef Q_OS_WIN
+    // Try to accuire a lock on the file and keep it until we done.
+    // If the file is locked, abort before we perform the move on the server
+    const QString fileName = propagator()->fullLocalPath(_item->_file);
+    const auto lockMode = propagator()->syncOptions().requiredLockMode();
+    m_fileLock = FileSystem::lockFile(fileName, lockMode);
+    if (!m_fileLock) {
+        emit propagator()->seenLockedFile(fileName, lockMode);
+        abortWithError(SyncFileItem::SoftError, tr("%1 the file is currently in use").arg(QDir::toNativeSeparators(fileName)));
+        return;
+    }
+#endif
+
     auto job = new MoveJob(propagator()->account(), source, destination, headers, this);
     _jobs.append(job);
     connect(job, &MoveJob::finishedSignal, this, &PropagateUploadFileNG::slotMoveJobFinished);
@@ -375,14 +398,6 @@ void PropagateUploadFileNG::startNextChunk()
     _currentChunkSize = qMin(propagator()->_chunkSize, _rangesToUpload.first().size);
 
     const QString fileName = propagator()->fullLocalPath(_item->_file);
-    // If the file is currently locked, we want to retry the sync
-    // when it becomes available again.
-    const auto lockMode = propagator()->syncOptions().requiredLockMode();
-    if (FileSystem::isFileLocked(fileName, lockMode)) {
-        emit propagator()->seenLockedFile(fileName, lockMode);
-        abortWithError(SyncFileItem::SoftError, tr("%1 the file is currently in use").arg(fileName));
-        return;
-    }
     auto device = std::unique_ptr<UploadDevice>(new UploadDevice(
             fileName, _currentChunkOffset, _currentChunkSize, &propagator()->_bandwidthManager));
     if (!device->open(QIODevice::ReadOnly)) {
@@ -488,7 +503,7 @@ void PropagateUploadFileNG::slotPutFinished()
     }
 
     // Check whether the file changed since discovery.
-    if (!FileSystem::verifyFileUnchanged(fullFilePath, _item->_size, _item->_modtime)) {
+    if (FileSystem::fileChanged(fullFilePath, _item->_size, _item->_modtime)) {
         propagator()->_anotherSyncNeeded = true;
         if (!_finished) {
             abortWithError(SyncFileItem::SoftError, tr("Local file changed during sync."));
